#!/usr/bin/env python
# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Format add-ons
==============

Format related functions and classes. Highlights: A TSV class, that helps
to work with tabular data.

Example:

    def run(self):
        with self.output().open('w') as output:
            output.write_tsv('Ubik', '1969', '67871286')

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

"""

from __future__ import unicode_literals

import collections
import functools
from builtins import str

import luigi
from gluish.utils import random_string, which
from six import string_types

__all__ = ['TSV']


def write_tsv(output_stream, *tup, **kwargs):
    """
    Write argument list in `tup` out as a tab-separeated row to the stream.
    """
    encoding = kwargs.get('encoding') or 'utf-8'
    value = u'\t'.join([s for s in tup]) + '\n'
    if encoding is None:
        if isinstance(value, string_types):
            output_stream.write(value.encode('utf-8'))
        else:
            output_stream.write(value)
    else:
        output_stream.write(value.encode(encoding))


def iter_tsv(input_stream, cols=None, encoding='utf-8'):
    """
    If a tuple is given in cols, use the elements as names to construct
    a namedtuple.

    Columns can be marked as ignored by using ``X`` or ``0`` as column name.

    Example (ignore the first four columns of a five column TSV):

    ::

        def run(self):
            with self.input().open() as handle:
                for row in handle.iter_tsv(cols=('X', 'X', 'X', 'X', 'iln')):
                    print(row.iln)
    """
    if cols:
        cols = [c if not c in ('x', 'X', 0, None) else random_string(length=5)
                for c in cols]
        Record = collections.namedtuple('Record', cols)
        for line in input_stream:
            yield Record._make(line.decode(encoding).rstrip('\n').split('\t'))
    else:
        for line in input_stream:
            yield tuple(line.decode(encoding).rstrip('\n').split('\t'))


class TSVFormat(luigi.format.Format):
    """
    A basic CSV/TSV format.
    Discussion: https://groups.google.com/forum/#!topic/luigi-user/F813st16xqw
    """

    def hdfs_reader(self, input_pipe):
        raise NotImplementedError()

    def hdfs_writer(self, output_pipe):
        raise NotImplementedError()

    def pipe_reader(self, input_pipe):
        input_pipe.iter_tsv = functools.partial(iter_tsv, input_pipe)
        return input_pipe

    def pipe_writer(self, output_pipe):
        output_pipe.write_tsv = functools.partial(write_tsv, output_pipe)
        return output_pipe


class GzipFormat(luigi.format.Format):
    """
    A gzip format, that upgrades itself to pigz, if it's installed.
    """
    input = 'bytes'
    output = 'bytes'

    def __init__(self, compression_level=None):
        self.compression_level = compression_level
        self.gzip = ["gzip"]
        self.gunzip = ["gunzip"]

        if which('pigz'):
            self.gzip = ["pigz"]
            self.gunzip = ["unpigz"]

    def pipe_reader(self, input_pipe):
        return luigi.format.InputPipeProcessWrapper(self.gunzip, input_pipe)

    def pipe_writer(self, output_pipe):
        args = self.gzip
        if self.compression_level is not None:
            args.append('-' + str(int(self.compression_level)))
        return luigi.format.OutputPipeProcessWrapper(args, output_pipe)

class ZstdFormat(luigi.format.Format):
    """
    The zstandard format.
    """
    input = 'bytes'
    output = 'bytes'

    def __init__(self, compression_level=None):
        self.compression_level = compression_level
        self.zstd = ["zstd"]
        self.unzstd = ["unzstd"]

    def pipe_reader(self, input_pipe):
        return luigi.format.InputPipeProcessWrapper(self.unzstd, input_pipe)

    def pipe_writer(self, output_pipe):
        args = self.zstd
        if self.compression_level is not None:
            args.append('-' + str(int(self.compression_level)))
        return luigi.format.OutputPipeProcessWrapper(args, output_pipe)

TSV = TSVFormat()
Gzip = GzipFormat()
Zstd = ZstdFormat()



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Various intervals.
"""

import datetime

__all__ = [
    'biweekly',
    'daily',
    'every_minute',
    'hourly',
    'monthly',
    'quarterly',
    'semiyearly',
    'weekly',
    'yearly',
]

def every_minute(dt=datetime.datetime.utcnow(), fmt=None):
    """
    Just pass on the given date.
    """
    date = datetime.datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, 1, 0, dt.tzinfo)
    if fmt is not None:
        return date.strftime(fmt)
    return date

def hourly(dt=datetime.datetime.utcnow(), fmt=None):
    """
    Get a new datetime object every hour.
    """
    date = datetime.datetime(dt.year, dt.month, dt.day, dt.hour, 1, 1, 0, dt.tzinfo)
    if fmt is not None:
        return date.strftime(fmt)
    return date

def daily(date=datetime.date.today()):
    """
    Just pass on the given date.
    """
    return date

def weekly(date=datetime.date.today()):
    """
    Weeks start are fixes at Monday for now.
    """
    return date - datetime.timedelta(days=date.weekday())

def biweekly(date=datetime.date.today()):
    """
    Every two weeks.
    """
    return datetime.date(date.year, date.month, 1 if date.day < 15 else 15)

def monthly(date=datetime.date.today()):
    """
    Take a date object and return the first day of the month.
    """
    return datetime.date(date.year, date.month, 1)

def quarterly(date=datetime.date.today()):
    """
    Fixed at: 1/1, 4/1, 7/1, 10/1.
    """
    return datetime.date(date.year, ((date.month - 1)//3) * 3 + 1, 1)

def semiyearly(date=datetime.date.today()):
    """
    Twice a year.
    """
    return datetime.date(date.year, 1 if date.month < 7 else 7, 1)

def yearly(date=datetime.date.today()):
    """
    Once a year.
    """
    return datetime.date(date.year, 1, 1)



#!/usr/bin/env python
# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

import datetime
import logging
import os

import luigi
import requests

from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout, which

__all__ = ['Executable', 'Available', 'GitCloneRepository', 'GitUpdateRepository', 'FillSolrIndex']

logger = logging.getLogger('gluish')


def service_is_up(service):
    """
    Return `False` if HTTP services returns status code != 200.
    """
    try:
        return requests.get(service).status_code == 200
    except:
        return False


def getfirstline(file, default):
    """
    Returns the first line of a file.
    """
    with open(file, 'rb') as fh:
        content = fh.readlines()
        if len(content) == 1:
            return content[0].decode('utf-8').strip('\n')

    return default


class chdir(object):
    """
    Change directory temporarily.
    """

    def __init__(self, path):
        self.wd = os.getcwd()
        self.dir = path

    def __enter__(self):
        os.chdir(self.dir)

    def __exit__(self, *args):
        os.chdir(self.wd)


class Executable(luigi.Task):
    """
    Checks, whether an external executable is available. This task will consider
    itself complete, only if the executable `name` is found in $PATH.
    """
    name = luigi.Parameter()
    message = luigi.Parameter(default="")

    def run(self):
        """ Only run if, task is not complete. """
        raise RuntimeError('external program required: %s (%s)' % (self.name, self.message))

    def complete(self):
        return which(self.name) is not None


class Available(luigi.Task):
    """
    Checks, whether an HTTP service is available or not. This task will consider
    itself complete, only if the HTTP service return a 200.
    """
    service = luigi.Parameter()
    message = luigi.Parameter(default="")

    def run(self):
        """ Only run if, task is not complete. """
        raise RuntimeError('HTTP service %s is not available (%s)' % (self.service, self.message))

    def complete(self):
        return service_is_up(self.service)


class GitCloneRepository(luigi.Task):
    """
    Checks, whether a certain directory already exists (that should contain a Git repository) - otherwise it will be cloned.
    """
    gitrepository = luigi.Parameter()
    repositorydirectory = luigi.Parameter()
    basedirectory = luigi.Parameter()

    def requires(self):
        return Executable(name='git')

    def run(self):
        self.output().makedirs()
        with chdir(str(self.basedirectory)):
            shellout("""git clone {gitrepository}""", gitrepository=self.gitrepository)

    def output(self):
        return luigi.LocalTarget(path=os.path.join(self.basedirectory, str(self.repositorydirectory)))


class GitUpdateRepository(luigi.Task):
    """
    Updates an existing Git repository
    """
    gitrepository = luigi.Parameter()
    repositorydirectory = luigi.Parameter()
    basedirectory = luigi.Parameter()
    branch = luigi.Parameter(default="master", significant=False)

    def requires(self):
        return [
            GitCloneRepository(gitrepository=self.gitrepository,
                               repositorydirectory=self.repositorydirectory,
                               basedirectory=self.basedirectory),
            Executable(name='git')
        ]

    def run(self):
        with chdir(str(self.output().path)):
            shellout("""git checkout {branch}""", branch=self.branch)
            shellout("""git pull origin {branch}""", branch=self.branch)

    def complete(self):
        if not self.output().exists():
            return False

        with chdir(str(self.output().path)):
            output = shellout("""git fetch origin {branch} > {output} 2>&1""", branch=self.branch)

            result = True

            with open(output, 'rb') as fh:
                content = fh.readlines()
                if len(content) >= 3:
                    result = False

            revparseoutput = shellout("""git rev-parse {branch} > {output} 2>&1""", branch=self.branch)
            originrevparseoutput = shellout("""git rev-parse origin/{branch} > {output} 2>&1""", branch=self.branch)

            revparse = getfirstline(revparseoutput, "0")
            originrevparse = getfirstline(originrevparseoutput, "1")

            if revparse != originrevparse:
                result = False

        return result

    def output(self):
        return luigi.LocalTarget(path=os.path.join(str(self.basedirectory), str(self.repositorydirectory)))


class FillSolrIndex(luigi.Task):
    # TODO: define proper complete criteria (?)
    # e.g. check, whether the amount of records that should be loaded into the index is index (if not, then index load is not successfully)
    """
    Loads processed data of a data package into a given Solr index (with help of solrbulk)
    """
    date = ClosestDateParameter(default=datetime.date.today())
    solruri = luigi.Parameter()
    solrcore = luigi.Parameter()
    purge = luigi.BoolParameter(default=False, significant=False)
    purgequery = luigi.Parameter(default="", significant=False)
    size = luigi.IntParameter(default=1000, significant=False)
    worker = luigi.IntParameter(default=2, significant=False)
    commitlimit = luigi.IntParameter(default=1000, significant=False)
    input = luigi.Parameter()
    taskdir = luigi.Parameter()
    outputfilename = luigi.Parameter()
    salt = luigi.Parameter()

    def determineprefix(self, purge=None, purgequery=None):
        solrbulk = 'solrbulk'

        if purge and purgequery is not None:
            return solrbulk + ' -purge -purge-query "' + purgequery + '"'
        if purge:
            return solrbulk + ' -purge'

        return solrbulk

    def requires(self):
        return [
            Available(service=self.solruri,
                      message="provide a running Solr, please"),
            Executable(name='solrbulk',
                       message='solrbulk command is missing on your system, you can, e.g., install it as a deb package on your Debian-based linux system (see https://github.com/miku/solrbulk#installation)'),
        ]

    def run(self):
        prefix = self.determineprefix(self.purge, self.purgequery)
        server = str(self.solruri) + str(self.solrcore)
        chunksize = self.size
        cores = self.worker
        inputpath = self.input
        commit = self.commitlimit
        output = shellout(
            """{prefix} -verbose -server {server} -size {size} -w {worker} -commit {commit} < {input}""",
            prefix=prefix,
            server=server,
            size=chunksize,
            worker=cores,
            commit=commit,
            input=inputpath)

        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=os.path.join(self.taskdir, str(self.outputfilename)))



# coding: utf-8
# pylint: disable=C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
A few utilities.
"""

import logging
import os
import random
import re
import string
import subprocess
import tempfile

from dateutil import relativedelta

__all__ = [
    'date_range',
    'random_string',
    'shellout',
]

logger = logging.getLogger('gluish')

def random_string(length=16):
    """
    Return a random string (upper and lowercase letters) of length `length`,
    defaults to 16.
    """
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))

def date_range(start_date, end_date, increment, period):
    """
    Generate `date` objects between `start_date` and `end_date` in `increment`
    `period` intervals.
    """
    next = start_date
    delta = relativedelta.relativedelta(**{period:increment})
    while next <= end_date:
        yield next
        next += delta

def which(program):
    """
    Search for program in PATH.
    """
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def shellout(template, preserve_whitespace=False, executable='/bin/bash',
             ignoremap=None, encoding=None, pipefail=True, **kwargs):
    """

    Takes a shell command template and executes it. The template must use the
    new (2.6+) format mini language. `kwargs` must contain any defined
    placeholder, only `output` is optional and will be autofilled with a
    temporary file if it used, but not specified explicitly.

    If `pipefail` is `False` no subshell environment will be spawned, where a
    failed pipe will cause an error as well. If `preserve_whitespace` is `True`,
    no whitespace normalization is performed. A custom shell executable name can
    be passed in `executable` and defaults to `/bin/bash`.

    Raises RuntimeError on nonzero exit codes. To ignore certain errors, pass a
    dictionary in `ignoremap`, with the error code to ignore as key and a string
    message as value.

    Simple template:

        wc -l < {input} > {output}

    Quoted curly braces:

        ps ax|awk '{{print $1}}' > {output}

    Usage with luigi:

        ...
        tmp = shellout('wc -l < {input} > {output}', input=self.input().path)
        luigi.LocalTarget(tmp).move(self.output().path)
        ....

    """
    if not 'output' in kwargs:
        kwargs.update({'output': tempfile.mkstemp(prefix='gluish-')[1]})
    if ignoremap is None:
        ignoremap = {}
    if encoding:
        command = template.decode(encoding).format(**kwargs)
    else:
        command = template.format(**kwargs)
    if not preserve_whitespace:
        command = re.sub('[ \t\n]+', ' ', command)
    if pipefail:
        command = '(set -o pipefail && %s)' % command
    logger.debug(command)
    code = subprocess.call([command], shell=True, executable=executable)
    if not code == 0:
        if code in ignoremap:
            logger.info("Ignoring error via ignoremap: %s" % ignoremap.get(code))
        else:
            logger.error('%s: %s' % (command, code))
            error = RuntimeError('%s exitcode: %s' % (command, code))
            error.code = code
            raise error
    return kwargs.get('output')



# coding: utf-8
# pylint: disable=F0401,E1101,E1103
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Default task
============

A default task, that covers file system layout.
"""

import datetime
import hashlib
import os
import tempfile

import luigi
from gluish.parameter import ClosestDateParameter

__all__ = [
    'BaseTask',
    'MockTask'
]

def is_closest_date_parameter(task, param_name):
    """ Return the parameter class of param_name on task. """
    for name, obj in task.get_params():
        if name == param_name:
            return hasattr(obj, 'use_closest_date')
    return False

def delistify(x):
    """ A basic slug version of a given parameter list. """
    if isinstance(x, list):
        x = [e.replace("'", "") for e in x]
        return '-'.join(sorted(x))
    return x

class BaseTask(luigi.Task):
    """
    A base task with a `path` method. BASE should be set to the root
    directory of all tasks. TAG is a shard for a group of related tasks.
    """
    BASE = os.environ.get('GLUISH_DATA', tempfile.gettempdir())
    TAG = 'default'

    def closest(self):
        """ Return the closest date for a given date.
        Defaults to the same date. """
        if not hasattr(self, 'date'):
            raise AttributeError('Task has no date attribute.')
        return self.date

    def effective_task_id(self):
        """ Replace date in task id with closest date. """
        params = self.param_kwargs
        if 'date' in params and is_closest_date_parameter(self, 'date'):
            params['date'] = self.closest()
            task_id_parts = sorted(['%s=%s' % (k, str(v)) for k, v in params.items()])
            return '%s(%s)' % (self.task_family, ', '.join(task_id_parts))
        else:
            return self.task_id

    def taskdir(self):
        """ Return the directory under which all artefacts are stored. """
        return os.path.join(self.BASE, self.TAG, self.task_family)

    def path(self, filename=None, ext='tsv', digest=False, shard=False, encoding='utf-8'):
        """
        Return the path for this class with a certain set of parameters.
        `ext` sets the extension of the file.
        If `hash` is true, the filename (w/o extenstion) will be hashed.
        If `shard` is true, the files are placed in shards, based on the first
        two chars of the filename (hashed).
        """
        if self.BASE is NotImplemented:
            raise RuntimeError('BASE directory must be set.')

        params = dict(self.get_params())

        if filename is None:
            parts = []

            for name, param in self.get_params():
                if not param.significant:
                    continue
                if name == 'date' and is_closest_date_parameter(self, 'date'):
                    parts.append('date-%s' % self.closest())
                    continue
                if hasattr(param, 'is_list') and param.is_list:
                    es = '-'.join([str(v) for v in getattr(self, name)])
                    parts.append('%s-%s' % (name, es))
                    continue
                
                val = getattr(self, name)

                if isinstance(val, datetime.datetime):
                    val = val.strftime('%Y-%m-%dT%H%M%S')
                elif isinstance(val, datetime.date):
                    val = val.strftime('%Y-%m-%d')
                
                parts.append('%s-%s' % (name, val))

            name = '-'.join(sorted(parts))
            if len(name) == 0:
                name = 'output'
            if digest:
                name = hashlib.sha1(name.encode(encoding)).hexdigest()
            if not ext:
                filename = '{fn}'.format(ext=ext, fn=name)
            else:
                filename = '{fn}.{ext}'.format(ext=ext, fn=name)
            if shard:
                prefix = hashlib.sha1(filename.encode(encoding)).hexdigest()[:2]
                return os.path.join(self.BASE, self.TAG, self.task_family, prefix, filename)

        return os.path.join(self.BASE, self.TAG, self.task_family, filename)

class MockTask(BaseTask):
    """ A mock task object. Read fixture from path and that's it. """
    fixture = luigi.Parameter()

    def content(self):
        """ Return the content of the file in path. """
        with open(self.fixture) as handle:
            return handle.read()

    def run(self):
        """ Just copy the fixture, so we have some output. """
        luigi.LocalTarget(path=self.fixture).copy(self.output().path)

    def output(self):
        """ Mock output. """
        return luigi.LocalTarget(path=self.path(digest=True))



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Some glue around luigi.

Provides a base class, that autogenerates its output filenames based on some
tag, classname and parameters.

Additionally, provide some smaller utilities, like a TSV format.
"""

from setuptools import setup

setup(name='gluish',
      version='0.2.17',
      description='Luigi helper.',
      url='https://github.com/miku/gluish',
      author='Martin Czygan',
      author_email='martin.czygan@gmail.com',
      packages=[
          'gluish',
      ],
      package_dir={'gluish': 'gluish'},
      install_requires=[
          'future>=0.16',
          'luigi>=1.0.20',
          'python-dateutil>=2.2',
          'pytz>=2014.4',
          'requests>=2.18.4',
          'six>=1.9.0',
      ],
      )
